#!/usr/bin/python
# -*- coding: utf-8 -*-

# This script reads the output of a MD run which stores on each row the
# cartesian coordinates of several atoms (specified by HACK_ATOMS file) and
# then:
#
# 1) prepares a SIMPRE input file as output for each row of the original
# HACK_ATOMS.out
# IMPORTANT: It is assumed that the last atom coordinates belong to the central
# atom
#
# 2) runs SIMPRE1.1 for eah of the prepared inputs as long as they use the same
# simpre.par file

from math import sqrt, acos, atan2, pi
from time import time
from collections import Counter
import os
import sys

def read_HACKfile(fname, id_atoms):
    '''
    The function reads a HACK_ATOMS.out file and the file with the identity of
    the atoms, and returns:
       - a list of lists, where each list contains the coordinates of natoms in
         a frame
       - a dictionary with identities of the atoms in the frames
    '''
    f = open(fname, 'r')
    framelist = [] #each list contained in the FL contains coords for the selected atoms
    for line in f:
        line = line.rstrip('\n').rstrip('\r')
        line = [ float(coord) for coord in line.split() ]
        framelist.append(line)
    f.close()
    natoms = len(framelist[0])/3
    dict_atoms = identify_atoms(id_atoms)
    return framelist, natoms, dict_atoms


def identify_atoms(fname):
    '''
    User must provide file with the identifier of each donor atom in each line,
    in the same order as they are presented in the HACK_ATOMS file. The format
    is R#:OD1, where R# is the name of the aminoacid in one-letter code, and
    after the colon the donor atom is tagged (to differentiate between 2 O in a
    carboxylate, for instance).  key=value  i='R#:OD1'
    '''
    dict_atoms = {}
    f = open(fname, 'r')
    i = 1
    for line in f:
        line = line.rstrip('\n').rstrip('\r')
        dict_atoms[i] = line
        i += 1
    return dict_atoms

def coord_origin(coords):
    '''
    Takes one list form the framelist and substracts the coordinates from the
    last atom (coord[-3:]) to all the other ligand coordinates.
    '''
    natoms = len(coords)/3
    c_at = coords[-3:]   #central atom coordinates, IF last atom is the METAL
    lig_atoms = [coords[j:j+3] for j in range(0, natoms*3 - 3, 3) ]
    trasl_atoms = []
    for atom in lig_atoms:
        x, y, z = atom[0], atom[1], atom[2]
        tx, ty, tz = x - c_at[0], y - c_at[1], z - c_at[2]
        trasl_atoms.append([tx, ty, tz])
    return trasl_atoms


def cartesian2spherical(coords):
    '''
    The argument given is a list of natoms lists, with the 3 coordinates with
    the origin of coordinates already traslated respect to the metal. The
    result is expressed in degrees instead of radians.
    '''
    sph_coords = []
    for atom in coords:
        x, y, z = atom[0], atom[1], atom[2]
        r = sqrt(x**2 + y**2 + z**2)
        t = acos(z/r)
        p = atan2(y, x)
        tg = t*360/(2*pi)
        pg = p*360/(2*pi)
        if tg < 0:
            tg += 360
        if pg < 0:
            pg += 360
        sph_coords.append([r, tg, pg])
    return sph_coords


def closer_atoms(coords, dict_atoms):
    '''
    It will sort in ascending order the donor atoms coordinates so that the
    atoms closer than 3 angstroms to the central atom are selected. Also the
    REC model is applied, so the radius reduction and the effective charge for
    each atom is assigned.
    '''
    #Add the atom index for each
    for i in range(len(coords)):
        coords[i].append(i+1)
    #Select only atoms closer than 3 angstrom
    sel = [coords[i] for i in range(len(coords)) if coords[i][0] <= 3.0 ]
    #Add the residue and oxigen identifiers, for instance ['D7', 'OD1']
    for i in range(len(sel)):
        id_atom = dict_atoms[sel[i][3]]
        id_atom = id_atom.split(':')
        sel[i].append(id_atom)
    #Count the number of oxygens belonging to same residue (1 or 2)
    freq = Counter([sel[i][4][0] for i in range(len(sel))])
    #Default values of reduced radius (rr) and effective charge (ec),
    #when only 1 oxygen per residue is interacting with the central atom.
    rr_ec = {'D3' : [0.895, 0.105], 'N5' : [1.350, 0.069], 'D7' :\
 [0.895, 0.105], 'W9' : [1.350, 0.069], 'E11' : [0.895, 0.105] , 'E14' :\
 [0.895, 0.105], 'WAT' : [1.350, 0.069]}
    final_coords = []
    for i in range(len(sel)):
        sel[i].append(freq[sel[i][4][0]])   #add frequencies to list
        sel[i].append(rr_ec[sel[i][4][0]])  #add defaults for rr & ec
        if sel[i][5] > 1:
            sel[i][6] = [1.100, 0.085]
        coords = [ sel[i][0] - sel[i][6][0], sel[i][1], sel[i][2], sel[i][6][1] ]
        final_coords.append(coords)
    return final_coords


def create_simpre_dat(coords, Ln, dict_atoms):
    # Tranlate the coords respect to the metal atom
    trasl_atoms = coord_origin(coords)
    # Tranform to spherical coordinates
    sph_coords = cartesian2spherical(trasl_atoms)
    # Select atoms closer than 3 angstroms and apply REC by doing the effective
    # radius reduction and assigning effective charges
    sel = closer_atoms(sph_coords, dict_atoms)
    natoms = len(sel)
    foutput = 'simpre.dat'
    f = open(foutput, 'w')
    s ='1=Ce3+, 2=Pr3+, 3=Nd3+, 4=Pm3+, 5=Sm3+, 6=Tb3+, 7=Dy3+, 8=Ho3+, 9=Er3+, 10=Tm3+, 11=Yb3+, 12=user\n'
    f.write(s)
    f.write('\n')
    f.write(' {0:>2}    !ion code (from 1 to 12, see above)\n'.format(Ln))
    f.write(' {0:>2}    !number of effective charges\n'.format(natoms))
    for i in range(len(sel)):
        r, t, p, q = sel[i][0], sel[i][1], sel[i][2], sel[i][3]
        f.write(' {0:>2} {1:12.7f} {2:12.7f} {3:12.7f} {4:12.7f}\n'.format(i+1, r, t, p, q))
    f.write(2*'\n')
    f.close()
    return sel



if __name__ == '__main__':
    print 80*'*'
    print "This program generates input files in the simpre.dat format using data\
 from \nHACK_ATOMS.out type files."
    print 80*'*'
    print
    fname = sys.argv[1]
    id_atoms = sys.argv[2]
#    fname = 'ATOMS_100000.out'
#    id_atoms = 'id_atoms.txt'
    ti = time()
    framelist, natoms, dict_atoms = read_HACKfile(fname, id_atoms)
    tf = time()
    print "The files %s and %s have been read in %.3f seconds." % (fname, id_atoms, tf-ti)
    nframes = len(framelist)
    print "The file %s has %i frames and coordinates for %i atoms." % (fname, nframes, natoms)
    ti = time()

###Loop for simpre.dat creation and simpre execution

    for i in range(nframes):
        fcoords = create_simpre_dat(framelist[i], 6, dict_atoms)
        os.system('gfortran simpre.f -llapack -lblas')
        os.system('./a.out')
        #Extract the last 26 lines (energy levels), take first column, remove blank lines, and store in a line
        os.system("tail -n 26 simpre.out  | awk '{print $1}' | sed  '/^[ \t]*$/d' | cut -f1 | paste -s >> out.txt")
#        print 80*'*'
#        for j in range(len(fcoords)):
#            r, t, p, q = fcoords[j][0], fcoords[j][1], fcoords[j][2], fcoords[j][3]
#            print ' {0:>2} {1:12.7f} {2:12.7f} {3:12.7f} {4:12.7f}'.format(j+1, r, t, p, q)
    tf = time()
    print '%i simpre.dat files for the %s file have been processed through \
SIMPRE1.1 in %.3f seconds' %(nframes, fname, tf-ti)
